/*
 * viosample_main.c
 * Sample driver for devglue
 *
 * Copyright (c) 2016 Socionext Inc.
 */
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/errno.h>
#include <linux/irqdomain.h>

#include <asm/devglue.h>

#define VIO_NR_REGS 5
static char *reg_names[VIO_NR_REGS] = { "voc_lvl", "voc_voc", "voc_sg2", "voc_sc", "voc_afbcd" };
static void __iomem *vptr[VIO_NR_REGS];

#define VIO_NR_IRQS 4
static char *irq_names[VIO_NR_IRQS] = { "voc_vout7", "voc_vout5", "voc_vout4", "voc_vout3" };
static unsigned int hwirq[VIO_NR_IRQS];

static irqreturn_t viosample_handler(int irq, void *dev_id)
{
	return IRQ_HANDLED;
}

/* initialize module */
static int __init viosample_init_module(void)
{
	int i, j, ret;

	for(i=0; i<VIO_NR_REGS; i++) {
		vptr[i] = devglue_ioremap(reg_names[i], DEVGLUE_MEMORY_UC);
		if (IS_ERR(vptr[i])) {
			ret = PTR_ERR(vptr[i]);
			pr_err("init: cannot register reg%d (%d)\n", i, ret);
			goto err_regs;
		}
	}

	for(i=0; i<VIO_NR_IRQS; i++) {
		if ((ret = devglue_request_irq_byname(&hwirq[i], viosample_handler, 0, irq_names[i], NULL)) < 0) {
			pr_err("init: cannot register irq%d (%d)\n", i, ret);
			goto err_irqs;
		}
	}

	return 0;

err_irqs:
	for(j=0; j<i; j++) {
		devglue_free_irq(hwirq[j], NULL);
	}
err_regs:
	for(j=0; j<i; j++) {
		devglue_iounmap(vptr[j]);
	}
	return ret;
}

/* clean up module */
static void __exit viosample_cleanup_module(void)
{
	int i;

	for(i=0; i<VIO_NR_REGS; i++) {
		devglue_iounmap(vptr[i]);
	}

	for(i=0; i<VIO_NR_IRQS; i++) {
		devglue_free_irq(hwirq[i], NULL);
	}

	return;
}

module_init(viosample_init_module);
module_exit(viosample_cleanup_module);

MODULE_AUTHOR("Socionext Inc.");
MODULE_LICENSE("Dual BSD/GPL");
MODULE_DESCRIPTION("viosample: sample driver for devglue");
